##
# DICOM.pl 
# Niigata Univ. Dent. / H.Nishiyama
# 初版　2006.9.5
# 最終更新　2017.11.29
# DICOMタグ情報の分解・出力系のサブルーチン群
# 2017.11.29: -e パラメータにて "-"の時に、拡張子を空欄とする仕様へ変更。

#--- 
sub usage {
	print "usage(in other perl program): require 'DICOM.pl'; \n";
	print "=====\n";
	print "This is library file for other program\n";
}

&usage;
exit;
#---

###
# chkDICOM(*FH)
# DICOM画像かどうかのチェックをするルーチン
# return  1; #--- big-endian, tm.
# return  2; #--- little-endian (通常の暗黙のDICOM形式）
# return  3; #--- DICOM file big-endian with VR
# return  4; #--- DICOM file little-endian with VR
# return  5; #--- DICOM file big-endian without VR
# return  6; #--- DICOM file little-endian witout VR
# return  8; #--- Realia file little-endian
# change  meta->data [3-6][3-6]
# 
# return -3; #--- not DICOM file
###
sub chkDICOM {
	local(*FH) = @_;
	read(FH, $head, 0x90) or (return -1); #--- cannot get 16x9 bytes from header
	seek FH,0,0 or (return -2); #--- cannot seek to head position
	$len = substr($head, 4, 4); #--- 4文字目から４文字を取り出し、長さとする。
	$chkval = unpack("N",$len); #--- 符号なしLONG(bigEndian)への変換
	print "chkval = ".$chkval."\n";
	if ($chkval == 4) { #--- 00 00 00 00 00 00 00 04 //--- [0000][0000]
		return  1; #--- big-endian, TM format. --> 2:LittleEndianへ変換される
	}
	$chkval = unpack("V",$len); #--- 符号なしLONG(littleEndian)への変換
	print "chkval = ".$chkval."\n";
	if ($chkval == 4) { #--- 08 00 00 00 04 00 00 00 //--- [0008][0000]
		return  2; #--- little-endian
	}
	elsif ($chkval == 30) { #--- 08 00 05 00
		if ((substr($head,0,4) cmp pack("cxcx",0x08,0x05)) == 0) {
			return 8; #--- Realia output little-endian
		}
	}
#	$null_0x80 = pack("x128"); #--- 00 ... 00 (128個、0x80個) //--- DICOM File Preamble
#	if ((substr($head,0,0x80) cmp $null_0x80) != 0) {
#		return -3; #--- not DICOM file
#	}
	if ((substr($head,0x80,4) cmp "DICM") != 0) {
		return -3; #--- not DICOM file
	}
	$ret= -3; #--- not DICOM file
	if ((substr($head,0x84,4) cmp pack("x3c",0x02)) == 0) { #--- 
		$to_L = "N"; #--- DICOM file big-endian //--- 通常は、あり得ない
		$ret = -3; #--- ファイルメタ情報は明示的VRリトルエンディアンと決められているので、エラーとする。
	}
	if ((substr($head,0x84,4) cmp pack("cx3",0x02)) == 0) {
		$to_L = "V";
		$ret = 4; #--- DICOM file little-endian, Normal Type with VR information
	}
	if ($ret >= 3) { #--- 128個のPreambleを持つDICOM file
		print "withPreamble\n";
		seek(FH,0x84,0) or (return -1);
	}
	return $ret; 
}

###
# 2bytes単位の画像データの上下反転、little，big-endian形式の相互反転用サブルーチン
# 通常、dicom 7FE0-0010のデータを対象とする。
###
#sub reverse2 ($) {
#	local($target) = @_;
#	my(@d);
#	@d = unpack($to_S."*",$target);
#	if ($to_S = "v") {
#		$target = pack("n*",@d);
#	}
#	else {
#		$target = pack("v*",@d);
#	}
#	return $target;
#}

sub reverse2 ($) {
	local($target) = @_;
	my($i,$c);
	for ($i = 0; $i < length($target); $i += 2) {
		$c = substr($target,$i,2);
		substr($target,$i,2) = reverse($c);
	}
	return $target;
}



#######
# 2bytes単位の画像データの上下反転を行い、同時に、ガントリー傾斜角補正（画像のシフト）を行う
# ガントリー傾斜に未対応のソフト用への変換ルーチン（ex. Realia）
# shift量はピクセル単位に換算後の値とする。R->L+、A->P+、F->H+
# dicom 7FE0-0010のデータを対象とする。
#######
sub reverse2shift {
	local($tos,$data,$width,$max_shift,$shift) = @_;
	my($height,$org_height,$i,$j,$s1,$s2,$v1,$v2,@workdata, $total_shift, $chk_point, $black_value, $val, $chk_count);
	@workdata = unpack($tos."*",$data);
	$data = pack("v*",@workdata);
	@workdata = unpack("s*",$data);
	# n->vの変換のみでは、符号付きデータの処理が不可能なので、一端、sにて符号付きshotに変換する
	#---
	$black_value = -2048;
	$chk_count = 100;
	$s1 = int($shift);
	$s2 = abs($shift - $s1);
	$s1 = abs($s1);
	$org_height = ($#workdata+1)/$width;
	$i = $chk_point = int($org_height/2);
	for ($j = 0; $j < 10; $j++) {
		print @workdata[$i * $width + $j],":";
	}
	
	print "\n","shift:",$s1,":",$s2,"\n";
	if ($shift > 0) {
		unshift(@workdata,($black_value) x ($s1*$width));
		$total_shift = $s1;
		if ($s2 != 0) {
			unshift(@workdata,($black_value) x $width);
			$height = ($#workdata+1) / $width;
			push(@workdata,($black_value) x $width);
			for ($i = 0; $i < $height; $i++) {
				for ($j = 0; $j < $width; $j++) {
					$pos = $i*$width + $j;
					$v1 = @workdata[$pos];
					$v2 = @workdata[$pos+$width];
					$val = $v1 * $s2 + $v2 * (1-$s2);
					@workdata[$pos] = int($val);
				}
			}
			splice(@workdata,-$width);
			$total_shift++;
		}
		unshift(@workdata,($black_value) x (($max_shift-$total_shift) * $width));
#		splice(@workdata, -(($#workdata+1)/$width - $org_heigt) * $width);
	}
	elsif ($shift < 0) {
		push(@workdata,($black_value) x ($s1*$width));
		$total_shift = $s1;
		if ($s2 != 0) {
			push(@workdata,($black_value) x $width);
			$height = ($#workdata+1) / $width;
			unshift(@workdata,($black_value) x $width);
			for ($i = $height; $i > 0; $i--) {
				for ($j = 0; $j < $width; $j++) {
					$pos = $i*$width + $j;
					$v1 = @workdata[$pos];
					$v2 = @workdata[$pos-$width];
					$val = $v1 * $s2 + $v2 * (1-$s2);
					@workdata[$pos] = int($val);
				}
			}
			splice(@workdata,0,$width);
			$total_shift++;
		}
		push(@workdata,($black_value) x (($max_shift-$total_shift) * $width));
#		splice(@workdata, 0, (($#workdata+1)/$width - $org_heigt) * $width);
	}
	#---
	$data = pack("v*", @workdata);
	return $data;
}
#######
#
# getNextItem($printFlag, *IN, $type, $tag1, $tag2, $length, $item; $vrLen, $vr)
# 	$printFlag:チェック用出力の制御フラグ
# 	*IN:入力ファイル
# 	$type:endianの違い
# 出力
# 	$tag1,$tag2,$vr,$length,$item,$vr
# 戻り値
#	>0:正常($lengthVal)
#	-1~:エラー箇所、VRは-5
#
#	SQ:	DLen: 有効
#		item: FFFE-E000、DLen: 有効
#		...
#	SQ: DLen: 0000
#		item: FFFE
#
#
#######

sub getNextItem {
	# *FH:file-descriptor, $fpos:file_pos, $
	local($printFlag,*FH,$type2) = @_; #
	$blockStart = (1==0);
	$type = ($type2 < 10)?$type2:int($type2/10);
	if ($type == 1 || $type == 3 || $type == 5) {
		$to_S = "n";
		$to_L = "N";
		$revf = 1;
	}
	elsif ($type == 2 || $type == 4 || $type == 6 || $type == 8) {
		$to_S = "v";
		$to_L = "V";
		$revf = 0;
	}
	if ($type == 3 || $type == 4) {
		$withVR = (1==1);
	}
	else {
		$withVR = (1==0);
	}
	$item = "";
	read(FH, $tag1, 2) == 2 or ( return -1); #get tag1
	$tagVal_1 = unpack($to_S,$tag1);
	if ($tagVal_1 != 2) {
		$type = $type2 % 10;
		if ($type == 1 || $type == 3 || $type == 5) {
			$to_S = "n";
			$to_L = "N";
			$revf = 1;
		}
		elsif ($type == 2 || $type == 4 || $type == 6 || $type == 8) {
			$to_S = "v";
			$to_L = "V";
			$revf = 0;
		}
		if ($type == 3 || $type == 4) {
			$withVR = (1==1);
		}
		else {
			$withVR = (1==0);
		}
		$tagVal_1 = unpack($to_S,$tag1);
	}
	if ($revf) {
		$tag1 = pack("v",$tagVal_1);
	}
	if ($printFlag) {
		printf "[%X:",$tagVal_1;
	}
	read(FH, $tag2, 2) == 2 or (return -2);
	$tagVal_2 = unpack($to_S,$tag2);
	if ($revf) {
		$tag2 = pack("v",$tagVal_2);
	}
	if ($printFlag) {
		printf "%X]",$tagVal_2;
	}
	if ($withVR && ($tagVal_1 != 0xFFFE)) { #--- with VR
		read(FH, $vr, 2) == 2 or (return -5);
		if ($printFlag) {
			print "[$vr]";
		}
		if (index("OB,OW,OF,SQ,UT,UN",$vr)%3 == 0) {
			$vrLen = 4;
			read(FH, $blank, 2) == 2 or (return -5);
			$vr .= $blank;
			read(FH, $length, 4) == 4 or (return -3);
			$lengthVal = unpack($to_L,$length);
			if ($revf) {
				$length = pack("V",$lengthVal);
			}
			if ($length eq pack("C4",0xff,0xff,0xff,0xff)) {
				$blockStart = (1==1);
			}
		}
		else {
			$vrLen = 2;
			read(FH, $length, 2) == 2 or (return -3);
			$lengthVal = unpack($to_S,$length);
			if ($revf) {
				$length = pack("v",$lengthVal);
			}
		}
		if ($printFlag) {
			print "($vrLen)";
		}
	}
	else { #--- without VR
		$vrLen = 0;
		read(FH, $length, 4) == 4 or (return -3);
		$lengthVal = unpack($to_L,$length);
		if ($revf) {
			$length = pack("V",$lengthVal);
		}
		if ($length eq pack("C4",0xff,0xff,0xff,0xff)) {
			$blockStart = (1==1);
		}
	}
	if ($printFlag) {
		print "$lengthVal:";
	}
	if ($blockStart) {
		if ($printFlag) {
			print ("[可変長タグ]\n");
		}
		$blockStart = (1==0);
		$item="";
		$item_part="";
	}
	else {
		read(FH, $item, $lengthVal) == $lengthVal or (return -4);
		#	if ($revf && $tagVal_1 == 0x7fe0 && $tagVal_2 == 0x10) {
		#		$item = reverse2($item);
		#	}
		if ($lengthVal > 256) {
			$item_part =  substr($item, 0, 256);
		}
		else {
			$item_part = $item;
		}
		if ($lengthVal == 2) {
			$num = unpack($to_S,$item_part);
			if ($printFlag) {
				printf "<%X>",$num;
			}
		}
		if ($lengthVal == 4) {
			$num = unpack($to_L,$item_part);
			if ($printFlag) {
				printf "<%X>",$num;
			}
		}
		$item_part =~ s/[\x00-\x19]/\./g;
		if ($printFlag) {
			print $item_part;
			print "\n";
		}
	}
	$_[3] = $tag1;
	$_[4] = $tag2;
	$_[5] = $length;
	$_[6] = $item;
	$_[7] = $vrLen;
	if ($withVR) {
		$_[8] = $vr;
	}
	else {
		$_[8] = "";
	}
	return($lengthVal);
}

##
# パラメータをチェックし、パラメータなら文字列を返す。
##

sub checkParam($) {
	local ($argv) = @_; 
	my ($c,$v);
	($v = shift(@$argv)) or (return 0);
#	print "arg:$v\n";
	$c = ord($v);
	if ($c == ord('-')) {
		$c = substr($v,1);
		$c =~ tr/A-Z/a-z/;
#		print "get param:$c\n";
		return $c;
	}
	else {
		unshift(@$argv,$v);
		return 0;
	}
}

###
# evalParam($argv, $_cv, $ext, $outdir, $printFlag, $printSecInfo, $sequential, $makeSubDir)
#
###

sub evalParam($) { # @$_cv
	local($argv, $_cv, $ext, $outdir, $printFlag, $outInfo) = @_;
	$_cc = 0;
	while ($v = checkParam($argv)) { 
		if ($v eq 'c') { ##--- C)onvert C)ommands
			$filename = shift(@$argv) or (return -2);
			open(IN, "<$filename") or (return -3);
			$fh = \*IN;
			while (defined(my $line = <$fh>)) {
	#			print "$line";
				chomp($line);
			next	if (length($line) <= 1 || ord($line) == ord("#"));
				$_cc++;
				$c = substr($line,0,1);
				$c =~ tr/A-Z/a-z/;
				substr($line,0,1) = $c;
				$c = substr($line,1,1);
				@_cvp = split(/$c/,$line);
				push(@$_cv, [@_cvp]);
			}
			print "get $_cc (total $#_cv) lines   for commands\n";
#---for chech
#			foreach $_cvp (@$_cv) {
#				@_cvpv = @{$_cvp};
#				print $_cvpv[0]."---\n";
#				foreach $v (@{$_cvp}) {
#					print $v.":";
#				}
#				print "\n";
#			}
#---
			close(IN);
		}
		elsif ($v eq 'e') { ##--- E)xtention　拡張子, ex. ".dcm"
			##---変換後の拡張子のセット（デフォルトで、".dcm"）
			$$ext = ".dcm";
			$$ext = shift(@$argv) or (return -2);
			if ($$ext eq "-") {
				$$ext = "";
			} 
			elsif (not ($$ext =~ /^\.[0-9a-zA-Z]+$/)) {
				$$ext = ".dcm";
			}
			print "ext:$$ext\n";
		}
		elsif ($v eq 'd') {
			##---変換後のファイルの保存先
			$$outdir =".";
			$$outdir = shift(@$argv) or (return -2);
			print "output dir:$$outdir\n";
		}
		elsif ($v eq 'p') {
			$$printFlag = (1==1);
			print "print trace data :$$printFlag\n";
		}
		elsif ($v eq 'i') {
			$$outInfo = (1==1);
			print "output info file :$$printFlag\n";
		}
		elsif ($v eq 's') {
			$$sequential = (1==1);
			print "sequential output for TAG, NOT sorted in numerical TAG order\n";
		}
		elsif ($v eq 'm') {
			$$makeSubDir = (1==1);
			print "make sub directory with series num\n";
		}
		else {
			return -3;
		}
	}
	return 1;
}

#@x = ('3','4','5','2','1','8');
#$xx = pack("v*",@x);
#$y = reverse2shift("v",$xx,3,2,1);
#print unpack("v*",$y);
1;