##
# getallitem.pl 
# Niigata Univ. Dent. / H.Nishiyama
# 初版　2006.9.5
# 更新　2014.4.15
# DICOM.plのサブルーチンを利用するDICOMタグ情報のテキスト形式での出力プログラム
# 画像データ部分などの大量のデータ部分については、簡易出力ないし未出力
# Preamble部の有無に対応、VRの有無に対応、可変長タグに部分的に対応
# 2021/09/18
# Windows10でのWSL/WSL2にて稼動するLinux（Ubuntu）のperlで動作するよう use lib '.';を追加。

use lib '.';
require 'DICOM.pl';

$directory = ".";
#--- 
sub usage {
	print "usage: perl getallitem.pl dicom_files\n";
	print "   dicom_files: dicom files\n";
	print "=====\n";
	print "output tag information of DICOM files.\n";
}


#---コマンド行パラメータが無い場合
if ($#ARGV < 0) {
	&usage;
	exit;
}

while ($filename = shift(@ARGV)) {
	open(IN, "<$filename") or die "open error:$!";
	binmode(IN);
	$_mode = chkDICOM(IN);
	if ($_mode == 1) {
		print "1:Big Endian(ex.TM format) suspected\n";
	}
	elsif ($_mode == 2) {
		print "2:Little Endian suspected\n";
	}
	elsif ($_mode == 3) {
		print "3:Big Endian suspected\n";
	}
	elsif ($_mode == 4) {
		print "4:Little Endian (GE format) suspected\n";
		$_withVR = (1==1);
	}
	elsif ($_mode == 8) {
		print "5:Little Endian (Realia format) suspected\n";
	}
	else { 
		die "not dicom or tm file:$!";
	}
	if ($_mode== 1 || $_mode== 3) {
		$_to_S = "n";
		$_to_L = "N";
	}
	elsif ($_mode== 2 || $_mode== 4 || $_mode== 8) {
		$_to_S = "v";
		$_to_L = "V";
	}

	$_ret = 0;
	@_drec=();
	%_table = ();
	$_preamble = "";
	$_printFlag = (1==1);
	if ($_mode == 4) { #--- メタ情報の生成
		$_preamble = pack("x128"); #--- 00 ... 00 (128個、0x80個) //--- DICOM File Preamble
		$_preamble .= "DICM";
	}
	while($_ret >= 0) {
		$_ret = getNextItem($_printFlag, *IN, $_mode, $_tag1, $_tag2, $_length, $_item, $_vrLen, $_vr);
		if ($_ret >= 0) {
			if ($_mode < 10) {
				if ((unpack($_to_S,$_tag1) == 2) && (unpack($_to_S,$_tag2) == 0x10)) {
					if ($_item eq "1.2.840.10008.1.2"."\x00") {
						$_mode = $_mode*10 + 6; #--- DICOM file little-endian without VR
						if ($_printFlag) {
							print "---> change to LE without VR format\n";
						}
					}
					elsif ($_item eq "1.2.840.10008.1.2.1"."\x00") {
						$_mode = $_mode*10 + 4; #--- DICOM file little-endian with VR
						if ($_printFlag) {
							print "---> change to LE with VR format\n";
						}
					}
				}
			}
			push(@_drec,[$_tag1,$_tag2,$_length,$_item,$_vrLen,$_vr]);
			$_table{$_tag1.$_tag2} = $#_drec;
#			$_lengVal = unpack($_to_L,$_length);
#			print "$#_drec---$_lengVal pushed\n";
		}
	}
	close(IN);
}
